signature BACK_TRACK =
  sig
    (* non-deterministic back tracking monad *)
    type 'a T

    val ret     : 'a -> 'a T                   (* unit *)
    val bnd     : 'a T -> ('a -> 'b T) -> 'b T (* bind *)
                                          
    val err     : 'a T                  (* fail *)
    val seq     : ('a T * 'b T) -> 'b T (* succeed if a and b succeed *)
    val or      : ('a T * 'a T) -> 'a T (* succeed if either succeed *)
    val cut     : 'a T -> 'a T          (* succeed once *)

    val map     : ('a -> 'b) -> 'a T -> 'b T 
    val fold    : ('a * 'b -> 'b T)  -> 'b T -> 'a T -> 'b T
    val join    : ('a T) T -> 'a T      (* dfs join *)
    val join'   : ('a T) T -> 'a T      (* bfs join *)

    val ans     : 'a T -> ('a * 'a T) option  (* get one answer *)
    val eval    : 'a T -> 'a list             (* get all answers *)

    (* useful functions  *)

    val all     : 'a T list -> 'a T (* == List.foldr seq err *)
    val one     : 'a T list -> 'a T (* == List.foldr or err *)
    val rets    : 'a list -> 'a T   (* == one o (List.map ret) *)

    val bndLst : 'a T list -> ('a list -> 'b T) -> 'b T
    val bndOpt  : 'a T option -> ('a option -> 'b T) -> 'b T
    val bndPair : ('a T * 'b T) -> ('a * 'b -> 'c T) -> 'c T

  end

structure BT :> BACK_TRACK =
  struct

    (* lazy sequence *)
    datatype 'a seq = S of unit -> ('a * 'a seq) option

    type 'a T = 'a seq
    
    val err  = S (fn () => NONE)
    fun ret x = S (fn () => SOME (x,err))
    fun ans (S f) = f ()

    (* lazy map *)
    fun map f m = 
      S (fn () =>
	 (case ans m of
	    NONE => NONE
	  | SOME (x,m') => SOME(f x,map f m')))

    (* dfs merge *)
    fun join xs =
	S (fn () =>
	   case ans xs of
	     NONE => NONE
	   | SOME (x,xs) => let
	       fun loop NONE = ans (join xs)
		 | loop (SOME(y,ys)) =
		 SOME(y,S (fn () => loop (ans ys)))
	     in loop (ans x)
	     end)

    (* bfs merge *)
    fun join' xs = let
      fun mkSeq [] = err
	| mkSeq (x::xs) = S (fn () => SOME (x,mkSeq xs))
      fun loop (m,q) = 
	  S (fn () => 
	     case (ans m,q) of
	       (NONE,[]) => NONE
	     | (NONE,q)  => ans (loop (mkSeq (List.rev q),[]))
	     | (SOME(m,xss),q)  => 
		 (case ans m of
		    NONE => ans (loop(xss,q))
		  | SOME(x,xs) => SOME(x,loop(xss,xs::q))))
    in loop(xs,[])
    end

    fun bnd m f =  let
      val seqs = map f m
    in join seqs
    end

    fun cut m =
      S (fn () =>
	 case ans m of
	   NONE => NONE
	 | SOME(x,xs) => SOME(x,err))

    fun fold f b a =
      S (fn () =>
	 case ans a of
	   NONE => (ans b)
	 | SOME(x,xs) =>
	     ans (bnd (fold f b xs) (fn xs => f(x,xs))))

    fun seq (x,y) = bnd x (fn _ => y)
    fun or (x,y) = join (S (fn () => SOME(x, (S (fn () => SOME(y,err))))))
    fun eval m =
      (case ans m of
	 NONE => []
       | SOME (x,xs) => x::(eval xs))
    fun one xs = List.foldr or err xs 
    fun all xs = List.foldr seq err xs
    fun rets xs = List.foldr (fn (x,xs) => or (ret x,xs)) err xs

    fun bndLst [] f =  f []
      | bndLst (x::xs) f =
      bnd x (fn x => bndLst xs (fn xs => (f (x::xs))))

    fun bndOpt NONE f =  f NONE
      | bndOpt (SOME x) f = bnd x (fn x => f (SOME x))
    fun bndPair (x,y) f =  bnd x (fn x => bnd y (fn y => f (x,y)))
	     
  end
