signature PRIM_CTXT = sig
  type ctxt 
    val eq : ctxt * ctxt -> bool

    val mk : {b:bool,
	      em:bool,
	      i:bool,
	      s:bool,
	      tt:bool,
	      u:word,
	      size:word option,
 	      color:SMLNG.color option} -> ctxt
    val toRec : ctxt -> {b:bool,
			 em:bool,
			 i:bool,
			 s:bool,
			 tt:bool,
			 u:word,
			 size:word option,
			 color:SMLNG.color option}
    structure Get : sig
      val b       : ctxt -> bool
      val em      : ctxt -> bool
      val i       : ctxt -> bool
      val s       : ctxt -> bool
      val tt      : ctxt -> bool
      val u       : ctxt -> word
      val size    : ctxt -> word option
      val color   : ctxt -> SMLNG.color option
    end
      
  end

structure WordCtxt :> PRIM_CTXT =
  struct
    open SMLNG
    type crec = {b:bool, (* bit 0 *)
		em:bool, (* bit 1 *)
		 i:bool, (* bit 2 *)
		 s:bool, (* bit 3 *)
	        tt:bool, (* bit 4 *)
		 u:word,  (* bits 5,6 *)
	      size:word option, (* bits 7,8,9,10  0 = NONE *)
      	     color:color option (* bits 11,12,13,14  0 = NONE *)}
    structure W = Word
    type ctxt = Word.word
    fun eq (c1:word,c2) = c1 = c2
    structure Get =
      struct
	fun b  (x) = W.>(W.andb(x,W.<<(0wx1,0w0)),0w0)
	fun em (x) = W.>(W.andb(x,W.<<(0wx1,0w1)),0w0)
	fun i  (x) = W.>(W.andb(x,W.<<(0wx1,0w2)),0w0)
	fun s  (x) = W.>(W.andb(x,W.<<(0wx1,0w3)),0w0)
	fun tt (x) = W.>(W.andb(x,W.<<(0wx1,0w4)),0w0)
	fun u  (x) = W.<<(W.andb(x,W.<<(0wx3,0w5)),0w5)
	fun size (x) = let
	  val x = W.<<(W.andb(x,W.<<(0wxF,0w7)),0w7)
	in if x = 0w0 then NONE
	   else SOME (W.-(x,0w1))
	end
	fun color(x) = 
	  (case W.<<(W.andb(x,W.<<(0wxF,0w11)),0w11) of
	     0w0 => NONE
	   | 0w1 => SOME Red
	   | 0w2 => SOME Green
	   | 0w3 => SOME Blue
	   | 0w4 => SOME Cyan
	   | 0w5 => SOME Magenta
	   | 0w6 => SOME Yellow
	   | 0w7 => SOME Black
	   | 0w8 => SOME White)
      end
    fun mk {b,em,i,s,tt,u,size,color} = let

      fun b2w true = 0w1
	| b2w false = 0w2
      fun u2w 0w0 = 0w0
	| u2w 0w1 = 0w1
	| u2w 0w2 = 0w2
	| u2w 0w3 = 0w3
	| u2w _ = 0w3
      fun size2w NONE = 0w0
	| size2w (SOME x) = W.+(W.min(x,0w9),0w1)

      fun color2w c  =
	(case c of
	    NONE => 0w0
	  | SOME Red    => 0w1
	  | SOME Green => 0w2 
	  | SOME Blue  => 0w3 
	  | SOME Cyan   => 0w4 
	  | SOME Magenta => 0w5
	  | SOME Yellow => 0w6
	  | SOME Black => 0w7 
	  | SOME White => 0w8) 
      val ctxt = 0w0
      val ctxt = W.orb(ctxt,W.<<(b2w b,0w0))
      val ctxt = W.orb(ctxt,W.<<(b2w em,0w1))
      val ctxt = W.orb(ctxt,W.<<(b2w i,0w2))
      val ctxt = W.orb(ctxt,W.<<(b2w s,0w3))
      val ctxt = W.orb(ctxt,W.<<(b2w tt,0w4))
      val ctxt = W.orb(ctxt,W.<<(u2w u,0w5))
      val ctxt = W.orb(ctxt,W.<<(size2w size,0w7))
      val ctxt = W.orb(ctxt,W.<<(color2w color,0w11))
    in ctxt
    end
    fun toRec x = {b=Get.b x,
		   em=Get.em x,
		   i=Get.i x,
		   s=Get.s x,
		   tt=Get.tt x,
		   u=Get.u x,
		   size=Get.size x,
		   color=Get.color x}
  end

structure SimpleCtxt :> PRIM_CTXT =
  struct
    open SMLNG
    type ctxt = {b:bool,
		em:bool,
		 i:bool,
		 s:bool,
	        tt:bool,
		 u:word,
	      size:word option,
 	      color:color option}
    fun eq (c1:ctxt,c2:ctxt) = c1 = c2
    structure Get =
      struct
	fun mkC x = x
	fun toC x = x
	fun b ({b=x,...}:ctxt) = x
	fun em ({em=x,...}:ctxt) = x
	fun i ({i=x,...}:ctxt) = x
	fun s ({s=x,...}:ctxt) = x
	fun tt ({tt=x,...}:ctxt) = x
	fun u ({u=x,...}:ctxt) = x
	fun size ({size=x,...}:ctxt) = x
	fun color ({color=x,...}:ctxt) = x
      end
    fun mk x = x
    fun toRec x = x
  end

structure Context :> CONTEXT =
  struct
    open SMLNG
    open SimpleCtxt

    fun upd (B,x) = let
      val {b,em,i,s,tt,u,size,color} = toRec x
    in mk {b=true,
	   em=em,
	   i=i,
	   s=s,
	   tt=tt,
	   u=u,
	   size=size,color=color}
    end
      | upd (EM,x) = let
       val {b,em,i,s,tt,u,size,color} = toRec x
	in mk {b=b,
	       em=if not s then not em else em,
	       i=i,
	       s=s,
	       tt=tt,
	       u=u,
	       size=size,color=color}
	end
      | upd (I,x) = let
	val {b,em,i,s,tt,u,size,color} = toRec x
	in mk {b=b,
	       em=em,
	       i=true,
	       s=s,
	       tt=tt,
	       u=u,
	       size=size,color=color}
	end
      | upd (PL,x ) = let 
	val {b,em,i,s,tt,u,size,color} = toRec x
	in mk {b=false,
	       em=false,
	       i=false,
	       s=false,
	       tt=false,
	       u=0w0,
	       size=size,color=color}
	end
      | upd (S,x) = let
	val {b,em,i,s,tt,u,size,color} = toRec x
	in mk {b=b,
	       em=false,
	       i=i,
	       s=true,
	       tt=tt,
	       u=u,
	       size=size,color=color}
	end
      | upd (TT,x) = let
	val {b,em,i,s,tt,u,size,color} = toRec x
	in mk {b=b,
	       em=em,
	       i=i,
	       s=s,
	       tt=true,
	       u=u,
	       size=size,color=color}
	end
      | upd (U,x) = let 
	  val {b,em,i,s,tt,u,size,color} = toRec x
	in mk {b=b,
	       em=em,
	       i=i,
	       s=s,
	       tt=tt,
	       u=if Word.<(u,0w3) then Word.+(u,0w1) else u,
	       size=size,color=color}
	end
      | upd (Size x,y) = let
	val {b,em,i,s,tt,u,size,color} = toRec y
	in mk {b=b,
	       em=em,
	       i=i,
	       s=s,
	       tt=tt,
	       u=u,
	       size=SOME x,color=color}
	end
      | upd (Color c,x) = let
	val {b,em,i,s,tt,u,size,color} = toRec x
	in mk {b=b,
	       em=em,
	       i=i,
	       s=s,
	       tt=tt,
	       u=u,
	       size=size,color=SOME c}
	end

    fun upd_ws x = let
      val {b,em,i,s,tt,u,size,color} = toRec x
    in mk {b=false,em=false,i=false,s=false,tt=false,u=u,size=NONE,
	   color=if u = 0w0 then (SOME White) else color}
    end

    val init =
      mk {b=false,em=false,i=false,s=false,tt=false,u=0w0,size=NONE,color=NONE}
    exception Bail

    fun reset (B,{b,em,i,s,tt,u,size,color}) =
      {b=false,em=em,i=i,s=s,tt=tt,u=u, size=size,color=color}
      | reset (EM,{b,em,i,s,tt,u,size,color}) =
      {b=b,em=false,i=i,s=s,tt=tt,u=u, size=size,color=color}
      | reset (I,{b,em,i,s,tt,u,size,color}) =
      {b=b,em=em,i=false,s=s,tt=tt,u=u, size=size,color=color}
      | reset (PL,{b,em,i,s,tt,u,size,color}) =
      {b=false,em=false,i=false,s=false,tt=false,u=0w0, size=size,color=color}
      | reset (S,{b,em,i,s,tt,u,size,color}) =
      {b=b,em=em,i=i,s=false,tt=tt,u=u, size=size,color=color}
      | reset (TT,{b,em,i,s,tt,u,size,color}) =
      {b=b,em=em,i=i,s=s,tt=false,u=u, size=size,color=color}
      | reset (U,{b,em,i,s,tt,u,size,color}) =
      {b=b,em=em,i=i,s=s,tt=tt,u=0w0, size=size,color=color}
      | reset (Size _,{b,em,i,s,tt,u,size,color}) =
      {b=b,em=em,i=i,s=s,tt=tt,u=u, size=NONE,color=color}
      | reset (Color _,{b,em,i,s,tt,u,size,color}) =
      {b=b,em=em,i=i,s=s,tt=tt,u=u, size=size,color=NONE}
    val reset = (fn (x,c) => mk (reset(x,toRec c)))
    fun reset_color x = reset(Color White,x)
    fun reset_size x = reset(Size 0w0,x)

    fun isNop (B,x) = (Get.b x) = true
      | isNop (EM,x) = if (Get.s x) then true else false
      | isNop (I,x) = (Get.i x) = true
      | isNop (PL,x) =   (Get.b x)  = false
		 andalso (Get.em x) = false
		 andalso (Get.i x)  = false
		 andalso (Get.tt x) = false
		 andalso (Get.s x)  = false
		 andalso (Get.u x)  = 0w0
      | isNop (S,x)       = (Get.s x) = true
      | isNop (TT,x)      = (Get.tt x) = true
      | isNop (U,x)       = Word.>=(Get.u x,0w3)
      | isNop (Size x,y)  = (Get.size y) = (SOME x)
      | isNop (Color x,y) = (Get.color y) = (SOME x)

    fun intersect (x,y) = let
      val b  = (Get.b x)  andalso (Get.b y)
      val em = (Get.em x) andalso (Get.em y)
      val i  = (Get.i x)  andalso (Get.i y)
      val s  = (Get.s x)  andalso (Get.s y)
      val tt = (Get.tt x) andalso (Get.tt y)
      val u  = Word.min(Get.u x,Get.u y)

      val (sx,sy) =  (Get.size x,Get.size y)
      val size   = if sx = sy then sx else NONE

      val (cx,cy) =  (Get.color x,Get.color y)
      val color  = if cx = cy then cx else NONE
    in mk {b=b,em=em,i=i,s=s,tt=tt,u=u,size=size,color=color}
    end

    fun append (x,y) = let
      val b  = (Get.b x)  orelse (Get.b y)
      val i  = (Get.i x)  orelse (Get.i y)
      val s  = (Get.s x)  orelse (Get.s y)
      val tt = (Get.tt x) orelse (Get.tt y)
      val em = if s then false
	       else not (Get.em x = Get.em y)
      val (u1,u2) =  (Get.u x,Get.u y)
      val u       = Word.min(0w3,Word.+(u1,u2))
      val (sx,sy) =  (Get.size x,Get.size y)
      val size   = case sx of
                     SOME _ => sx
		      | NONE => sy
      val (cx,cy) =  (Get.color x,Get.color y)
      val color   = case cx of
                     SOME _ => cx
		      | NONE => cy
    in mk {b=b,em=em,i=i,s=s,tt=tt,u=u,size=size,color=color}
    end

    fun difference (x,y) = let
      val b  = (Get.b x)  andalso not (Get.b y)
      val em = (Get.em x) andalso not (Get.em y)
      val i  = (Get.i x)  andalso not (Get.i y)
      val s  = (Get.s x)  andalso not (Get.s y)
      val tt = (Get.tt x) andalso not (Get.tt y)
      val (u1,u2) =  (Get.u x,Get.u y)
      val u       = if u1 > u2 then u1 - u2 else 0w0
      val (sx,sy) =  (Get.size x,Get.size y)
      val size   = if sx = sy then NONE else sx
      val (cx,cy) =  (Get.color x,Get.color y)
      val color  = if cx = cy then NONE else cx
    in mk {b=b,em=em,i=i,s=s,tt=tt,u=u,size=size,color=color}
    end

    fun toTags x = let
      val {b,em,i,s,tt,u,size,color} = toRec x
      val tags = []
      val tags = if b  then B::tags else tags
      val tags = if em then EM::tags else tags
      val tags = if i  then I::tags else tags
      val tags = if s  then S::tags else tags
      val tags = if tt then TT::tags else tags
      val tags =
	case u of
	  0w0 => tags
	| 0w1 => U::tags
	| 0w2 => U::U::tags
	| 0w3 => U::U::U::tags
	| _ => raise (Fail "bug")
      val tags =
	case size of
	  NONE => tags
	| SOME size => (Size size)::tags
      val tags =
	case color of
	  NONE => tags
	| SOME color => (Color color)::tags
    in tags
    end


    fun fold_with_ctxt {kSeq,kTxt,kWs,kNd} ctxt d = let
      fun f (ctxt,Seq(d1,d2)) = kSeq(f (ctxt,d1),f (ctxt,d2))
	| f (ctxt,Txt s) = kTxt (ctxt,s)
	| f (ctxt,Ws s) = kWs (upd_ws ctxt,s)
	| f (ctxt,Nd (t,d)) = kNd(ctxt,t,f (upd(t,ctxt),d))
    in f(ctxt,d)
    end

  end
