signature EM_ELIMINATION = sig
  include PASS
  val dbg : bool ref
  val em_tests : string list
end
structure EMElimination :> EM_ELIMINATION =
  struct
    structure T = SMLNG
    structure C = Context
    val {dsay,derr,ddoc,dbg} = Debug.mk ()
    val name = "em-elimination"
    type cost =  {t:int,f:int}
    fun add_cost (x:cost,y:cost) =
      {t = (#t x) + (#t y), f = (#f x) + (#f y)}
    val zero_cost = {t=0,f=0}
    (* assumes all redundant EMs removed *)

    fun mkcost true  = {t=0,f=9}
      | mkcost false = {t=9,f=0}

    (* ensure node x has context em' when of parent has context em *)
    fun ensure_em x em (SOME (em':bool)) =
      if em = em' then x else T.Nd(T.EM,x)
      | ensure_em x em NONE = x 

    fun go doc = let
      fun kSeq((x,cx),(y,cy)) = let
	val {t=xt,f=xf} = cx
	val {t=yt,f=yf} = cy
	  (* compute four costs *)
         (* p x y *)           
	val T_t_t = (xt + yt + 0,(true,true))
	val T_t_f = (xt + yf + 9,(true,false))
	val T_f_t = (xf + yt + 9,(false,true))
	val T_f_f = (xf + yf + 9,(false,false))

	val F_t_t = (xt + yt + 9,(true,true))
	val F_t_f = (xt + yf + 9,(true,false))
	val F_f_t = (xf + yt + 9,(false,true))
	val F_f_f = (xf + yf + 0,(false,false))
	  
	fun min(x as (xc,(xl,xr)),y as (yc,(yl,yr))) =
	  (case (Int.compare(xc,yc)) of
	    LESS => x
	  | GREATER => y
	  | EQUAL => x)
	val (t,T) = List.foldl min T_t_t [T_t_t,T_t_f,T_f_t,T_f_f] 
	val (f,F) = List.foldl min T_t_t [F_t_t,F_t_f,F_f_t,F_f_f] 
	val cost = {t=t,f=f}
	(* val sum = add_cost (cx,cy) *)
	fun mk _ f NONE = f NONE
	  | mk m f (SOME (x:bool)) = if x = m then (f (SOME m)) 
			      else T.Nd(T.EM,f (SOME m))
	fun mks _ NONE = T.Seq(x NONE,y NONE)
	  | mks ((l,r),_) (p as (SOME true)) = 
	  if l = r then mk l (fn v => T.Seq(x v,y v)) p
	  else T.Seq(mk l x p,mk r y p)
	  | mks (_,(l,r)) (p as (SOME false)) = 
	  if l = r then mk l (fn v => T.Seq(x v,y v)) p
	  else T.Seq(mk l x p,mk r y p)
      in (mks (T,F),cost)
      end

      fun kTxt(c,"") = ((fn _ => T.Txt ""),zero_cost)
	| kTxt(c,s) = let
	val em = C.Get.em c
      in (ensure_em (T.Txt s) em,mkcost em)
      end
      fun kWs(c,s)  = ((fn _ => (T.Ws s)),zero_cost)
      fun kNd(c,T.EM,(d,cd)) = let
	val parent_em = C.Get.em c
	val child_em = not parent_em
	val {t,f} = cd
	val min_set = (t < f) 
	val cost = 	    
	  case (min_set,parent_em) of
	    (true,false) => {t=t,f=f+9}
	  | (false,true) => {t=t+9,f=f}
	  | (_,_) => cd
	fun mk NONE = (d NONE)
	  | mk (SOME x) = if x = min_set then (d (SOME min_set))
			  else T.Nd(T.EM,d (SOME min_set))
      in (mk,cost)
      end
	| kNd(c,T.PL,(d,cd)) = let
	    val {t,f} = cd
	    val set_em = 
	      if (t + 9) = f then false
	      else (t + 9 < f)
	    fun mk v =
	      T.Nd(T.PL,if set_em then T.Nd(T.EM,d (SOME true))
			else (d (SOME false)))
	  in (mk,zero_cost)
	  end
	| kNd(c,T.S,(d,cd)) = let
	  in ((fn v => T.Nd(T.S,d NONE)),zero_cost)
	  end
	| kNd(c,t,(d,cd)) = let
	  in ((fn v => (T.Nd(t,d v))),cd)
	  end
      val (mk,{t,f}) =
	Context.fold_with_ctxt {kSeq=kSeq, kTxt=kTxt, kWs=kWs, kNd=kNd} 
	Context.init doc
      val doc = if (t + 9) = f then (mk (SOME false))
		else if (t + 9) < f then T.Nd(T.EM,(mk (SOME true)))
		     else (mk (SOME false))
    in doc
    end
    val em_tests =
      ["<S> foo <EM> bar </EM> </S> <EM> <EM> foo </EM> </EM>",
       "<EM>1</EM><EM>1</EM><EM>1</EM><EM>1</EM><EM>1</EM>",
       "<EM><EM>1</EM></EM>",
       "<EM><EM>1<EM>2</EM></EM></EM>",
       "<S><EM>1</EM></S>",
       "<EM><S><EM>1</EM></S></EM>",
       "<EM><EM><S>1<EM>2</EM></S></EM></EM>",
       "<EM><EM>1</EM><EM>2</EM>3</EM>",
       "<EM>1</EM><EM>2</EM><EM>3</EM>4",
       "<PL>1</PL>",
       "<EM><PL>1</PL></EM>",
       "<EM><PL>1</PL>2</EM>",
       "<PL><EM>1</EM></PL>",
       "<PL><EM><EM>1</EM></EM></PL>",
       "<PL><EM>1</EM>2</PL>",
       "<EM><PL>1<EM>2</EM>3</PL>4</EM>"]
    val tests = Tests.mk name false go em_tests
  end
