(* intermediate SML/NG *)
(* optimization must be done in the following order
   canon -> pl_shortcut -> sharing_opt -> color_opt -> size_opt *)
signature INTER_NG =  sig
  include PASS

  datatype inter_ng =
    Seq of inter_ng * inter_ng
  | Txt of string * bool
  | Nd  of (Context.ctxt * inter_ng)

  val fromDoc    : SMLNG.doc -> inter_ng
  val toDoc      : inter_ng -> SMLNG.doc

  val canon      : inter_ng -> inter_ng
  val pl_shortcut: inter_ng -> inter_ng
  val sharing_opt: inter_ng -> inter_ng
  val color_opt  : inter_ng -> inter_ng
  val size_opt   : inter_ng -> inter_ng

  val dbg_out  : inter_ng -> unit
  val dbg : bool ref
  val id_tests : unit -> unit
end

structure InterNG :> INTER_NG =  struct
  datatype inter_ng =
    Seq of inter_ng * inter_ng
  | Txt of string * bool
  | Nd  of (Context.ctxt * inter_ng)

  val {dsay,derr,ddoc,dbg} = Debug.mk ()

  structure C = struct
    open Context
    open Context.Get
  end
  structure T = SMLNG

  val ctxt_and = C.intersect
  val ctxt_sub = C.difference

  (* pretty printer *)
  fun dbg_out ng = () (*let
    fun f (Seq(d1,d2)) = ((*dsay "<sq>";*)f d1;f d2(*;dsay "</sq>"*))
      | f (Txt(s,_)) = dsay s
      | f (Nd (c,d)) =
       let
	 val tags = C.toTags c
       in
(*          dsay "<nd>";*)
	  dsay (foldl (fn (x,y) => "<"^(SMLNG.tag2str x) ^ ">" ^ y) "" tags);
          f d;
 	  dsay (foldl (fn (x,y) => "</"^(SMLNG.tag2str x) ^ ">" ^ y) "" (List.rev tags))
(*	  dsay "</nd>"*)
       end
    in f ng;
       dsay "\n"
    end*)

  fun delete_empty_context (Nd(c,t)) =
      if Context.eq(c,Context.init) then delete_empty_context (t)
      else Nd(c, (delete_empty_context t))
    | delete_empty_context (Seq(t1,t2)) = Seq(delete_empty_context t1,
					      delete_empty_context t2)
    | delete_empty_context t = t

  fun delete_empty_string t = let
    fun loop (Seq(t1,t2)) =
      (case (loop t1, loop t2) of
	  (SOME t1', SOME t2') => SOME(Seq(t1',t2'))
	| (SOME t1', NONE) => SOME t1'
	| (NONE, SOME t2') => SOME t2'
	| (NONE, NONE) => NONE)
      | loop (Nd(c,t)) =
       (case loop t of
	   SOME t' => SOME(Nd(c,t'))
           | NONE => NONE)
      | loop (Txt(c,b)) =
       if c="" then NONE else SOME(Txt(c,b))
  in case loop t of
       SOME t' => t'
       | NONE => Txt("",true)
  end

  fun combine_context (Nd(c1,Nd(c2,t))) =
     Nd(C.append(c1,c2),combine_context t)
    | combine_context (Nd(c1,Seq(t1,t2))) =
     Nd(c1,Seq(combine_context t1, combine_context t2))
    | combine_context (Seq(t1,t2)) =
     Seq(combine_context t1, combine_context t2)
    | combine_context t = t

  fun getMax l =
     let
	fun max [x] n m = (x, n)
	  | max (x::y::xs) n m = if (x > y) then max (x::xs) n (m+1)
				 else max (y::xs) m (m+1)
	  | max [] _ _ = (0,0)
     in max l 1 2
     end

  fun gain_by_sharing(c) =
     let
	fun conv b = if b then 1 else 0
	val vsum = (conv (C.b c)) * 7 +
	           (conv (C.em c)) * 9 +
		   (conv (C.i c)) * 7 +
		   (conv (C.s c)) * 7 +
		   (conv (C.tt c)) * 9+
		   (Word.toInt (C.u c)) * 7 +
		   (case (C.size c) of SOME _ => 7 | NONE => 0) +
		   (case (C.color c) of SOME _ => 7 | NONE => 0)
     in vsum end

  fun pl_shortcut (ng) = let
    val font_ctxt1 =
       foldl C.upd C.init [T.B,T.EM,T.I,T.TT,T.U,T.U,T.U]
    val font_ctxt2 =
       foldl C.upd C.init [T.B,T.S,T.I,T.TT,T.U,T.U,T.U]

    (* a hack to add <PL> tag *)
    fun add_pl t = Seq(Txt("<PL>",true),Seq(t,Txt("</PL>",true)))

    fun st_win(c1,c2,c3) = let
      (* gain by shortcut PL *)
      val c1_and_c3 = ctxt_and(c1,c3)
      val c1c3_and_font =
	 if (C.em c1_and_c3) then ctxt_and(c1_and_c3,font_ctxt1)
	 else ctxt_and(c1_and_c3,font_ctxt2)
      val gain_sc = gain_by_sharing(c1c3_and_font) - 9

      (* another option: extract the common part of c1,c2,c3 *)
      val c1c2c3 = ctxt_and(c1_and_c3,c2)
      val gain' = gain_by_sharing(c1c2c3)

(*      val _ = dbg_out (Nd(c3,Txt("___c3",true)))
      val _ = dbg_out (Nd(c1c3_and_font,Txt("___c1c3font",true)))
*)
    in (gain_sc > gain', c1c3_and_font) end

    fun loop (Seq(Nd(c1,t1),(l as Seq(Nd(c2,t2),Seq(Nd(c3,t3),t4))))) =
      let val (win,common) = st_win(c1,c2,c3) in
        if win then
	   let
	      val cl1 = ctxt_sub(c1,common)
	      val cl3 = ctxt_sub(c3,common)
	   in loop(Seq(Nd(common,
			  Seq(Nd(cl1,t1),
			      Seq(add_pl(Nd(c2,t2)),Nd(cl3,t3)))),
		       t4))
	   end
	else Seq(Nd(c1,t1),loop l)
      end
      | loop (Seq(Nd(c1,t1),(l as Seq(Nd(c2,t2),Seq(_,t4))))) =
        Seq(Nd(c1,t1),loop l)
      | loop (Seq(Nd(c1,t1),Seq(t2,(l as Seq(Nd(c3,t3),t4))))) =
      let val (win,common) = st_win(c1,C.init,c3) in
        if win then
	   let
	      val cl1 = ctxt_sub(c1,common)
	      val cl3 = ctxt_sub(c3,common)
	   in loop(Seq(Nd(common,
			  Seq(Nd(cl1,t1),
			      Seq(add_pl(t2),Nd(cl3,t3)))),
		       t4))
	   end
	else Seq(Nd(c1,t1),Seq(t2,loop l))
      end
      | loop (l as (Seq(Nd(c1,t1),Seq(Nd(c2,t2),Nd(c3,t3))))) =
      let val (win,common) = st_win(c1,c2,c3) in
        if win then
	   let
	      val cl1 = ctxt_sub(c1,common)
	      val cl3 = ctxt_sub(c3,common)
	   in Nd(common,
		 Seq(Nd(cl1,t1),
		     Seq(add_pl(Nd(c2,t2)),Nd(cl3,t3))))
	   end
	else l
      end
      | loop (l as (Seq(Nd(c1,t1),Seq(Nd(c2,t2),Txt _)))) = l
      | loop (l as (Seq(Nd(c1,t1),Seq(t2,Nd(c3,t3))))) =
      let val (win,common) = st_win(c1,C.init,c3) in
        if win then
	   let
	      val cl1 = ctxt_sub(c1,common)
	      val cl3 = ctxt_sub(c3,common)
	   in Nd(common,
		 Seq(Nd(cl1,t1),
		     Seq(add_pl(t2),Nd(cl3,t3))))
	   end
	else l
      end
      | loop (l as (Seq(Nd(c1,t1),_))) = l
      | loop (Seq(t,t1)) = Seq(t,loop t1)
      | loop l = l

  in loop ng end

  fun sharing_opt (ng as (Seq(t1,t2))) = let
    fun get_weight_list (Seq(Nd(c1,_), tail as (Seq(Nd(c2,_) ,l)))) =
          gain_by_sharing(ctxt_and(c1,c2)) :: get_weight_list tail
    |   get_weight_list (Seq(Nd(c1,_), tail as (Seq(_ ,l)))) =
          0 :: 0 :: get_weight_list l
    |   get_weight_list (Seq(Nd(c1,_),Nd(c2,_))) = [gain_by_sharing(ctxt_and(c1,c2))]
    |   get_weight_list (Seq(Nd(c1,_),Txt(_))) = [0]
    |   get_weight_list (Seq(_,l)) = 0 :: get_weight_list l
    |   get_weight_list _ = [0]
    val wl=get_weight_list ng

   val (maxv,ps) = getMax wl

   fun transform (Seq (Nd(c1, t1), Seq (Nd(c2,t2), xs)), 1) =
      let
	 val c1_and_c2 = ctxt_and (c1, c2)
	 val c1_sub_c1c2 = ctxt_sub (c1, c1_and_c2)
	 val c2_sub_c1c2 = ctxt_sub (c2, c1_and_c2)
(*         val _ = dsay("\n")
         val _ = dbg_out (Nd(c1,Txt("___c1",true)))
         val _ = dbg_out (Nd(c2,Txt("___c2",true)))*)
	 val newNode = Nd (c1_and_c2, Seq(Nd(c1_sub_c1c2, t1),
					  Nd(c2_sub_c1c2, t2)))
      in Seq (newNode, xs)
      end
     | transform (Seq (Nd(c1, t1), (Nd(c2,t2))), 1) =
      let
	 val c1_and_c2 = ctxt_and (c1, c2)
	 val c1_sub_c1c2 = ctxt_sub (c1, c1_and_c2)
	 val c2_sub_c1c2 = ctxt_sub (c2, c1_and_c2)
	 val newNode = Nd (c1_and_c2, Seq(Nd(c1_sub_c1c2, t1),
					  Nd(c2_sub_c1c2, t2)))
      in newNode
      end
     | transform (Seq (x,xs), n) = Seq (x, transform (xs, n-1))
     | transform _ = raise (derr "impossible\n")

       in
	  if maxv > 0 then sharing_opt(transform(ng, ps))
	  else Seq(sharing_opt t1, sharing_opt t2)
       end
    | sharing_opt (Nd(c,t)) = Nd(c,sharing_opt t)
    | sharing_opt (Txt(s,b)) = Txt(s,b)

  fun find_longest_pair (getf,cl,t) = let
    fun loop(Seq(Nd(ctxt,_),l),dep) =
       (case getf ctxt of
	   SOME c =>
	      if c=cl then
		 let
		    val (dep1,rt1) = loop(l,dep+1)
		 in if dep>dep1 then (dep,SOME l) else (dep1,rt1) end
	      else loop(l,dep+1)
	 | NONE  => (0,NONE))
      | loop (Nd(ctxt,_),dep) =
       (case getf ctxt of
	   SOME c =>
	      if c=cl then (dep,NONE) else (0,NONE)
	 | NONE  => (0,NONE))
      | loop _ = (0,NONE)
  in loop(t,1) end

  fun color_size_sharing (getf, resetf, ctxt_cons, t) = let
    fun loop (t as (Seq(Nd(ctxt,t1),l))) =
     (case getf ctxt of
	SOME c =>
          let
	     val (nextpair,rt) = find_longest_pair(getf,c,l)

	     fun transform (Seq(Nd(ctxt,t1),_),1) =
               (case getf ctxt of
                 SOME c1 => if c = c1 then
  		              Nd(resetf ctxt,t1)
			    else Nd(ctxt,t1)
	          | NONE => raise (derr "impossible"))
	       | transform (Seq(Nd(ctxt,t1),l),n) =
               (case getf ctxt of
                 SOME c1 => if c = c1 then
		    Seq(Nd(resetf ctxt, t1), transform(l,n-1))
			    else Seq(Nd(ctxt,t1),transform(l,n-1))
	          | NONE => raise (derr "impossible"))
	       | transform (Nd(ctxt,t1),n) =
               (case getf ctxt of
                 SOME c1 => if c = c1 then
		            Nd(resetf ctxt, t1)
			    else Nd(ctxt,t1)
	          | NONE => raise (derr "impossible"))
	       | transform _ = raise (derr "impossible")

	  in if nextpair>0 then
             case rt of
		SOME t =>
         	     Seq(Nd(ctxt_cons(c),
			    Seq(Nd(resetf ctxt,loop(t1)),
				loop(transform(l,nextpair)))),
			 loop(t))
              | NONE =>  Nd(ctxt_cons(c),
			    Seq(Nd(resetf ctxt,loop(t1)),
				loop(transform(l,nextpair))))
	     else Seq(Nd(ctxt,loop(t1)), loop(l))
	  end
        | NONE => Seq(Nd(ctxt,loop(t1)), loop(l)))
    | loop (Seq(t1,l)) =
	 Seq(loop(t1), loop(l))
    | loop t = t
  in loop t end

  fun canon ng = let
    val txt_list = ref []
    fun f(c,Seq(Txt("<PL>",_),Seq(t1,Txt("</PL>",_)))) =
        f(C.upd(T.PL,c),t1)
      | f(c,Seq(d1,d2)) = (f(c,d2);f(c,d1))
      | f(c,Txt(s,b)) = let
          val lt = !txt_list
        in txt_list:= (c,s,b) :: lt end
      | f(c,Nd(c',d)) = let
	  val newc = C.append(c',c)
        in f(newc,d) end
    fun form_seq [] = Txt("",true)
      | form_seq [(c,s,b)] = Nd(c,Txt(s,b))
      | form_seq ((c,s,b)::xs) = Seq(Nd(c,Txt(s,b)),form_seq xs)
  in f(C.init,ng); form_seq(!txt_list)
  end

  fun fromDoc (T.Seq(d1, d2)) = Seq(fromDoc d1, fromDoc d2)
    | fromDoc (T.Txt s) = (*(dsay (s^"\n");*)Txt(s,true)
    | fromDoc (T.Ws s) = (*(dsay (s^"\n");*)Txt(s,false)
    | fromDoc (T.Nd(tg,d)) =
        if tg = T.PL then
	  Seq(Txt("<PL>",true),Seq(fromDoc d,Txt("</PL>",true)))
	else Nd(C.upd(tg,C.init),fromDoc d)

  fun toDoc x = let
    (* the first case is a hack to pl tag *)
    fun f(Seq(Txt("<PL>",_),Seq(d,Txt("</PL>",_)))) =
        T.Nd(T.PL,f d)
      | f(Seq(d1,d2)) = T.Seq(f d1,f d2)
      | f(Txt(s,true)) = T.Txt s
      | f(Txt(s,false)) = T.Ws s
      | f(Nd(c',d)) = let
	  val tags = C.toTags c'
	fun mkNode [] = f d
	  | mkNode (x::xs) = T.Nd(x,mkNode xs)
	in mkNode tags
	end
  in f x
  end

  val name = "inter-ng"
  fun color_opt d =
    color_size_sharing(C.color,C.reset_color,
		       (fn c => C.upd(T.Color c,C.init)),d)
  fun size_opt d =
    color_size_sharing(C.size,C.reset_size,
			       (fn s => C.upd(T.Size s,C.init)),d)
  fun go d = let
    val t = fromDoc d
    val t = delete_empty_string t
    val sz = T.size d

    val (opt1,d1) = if (sz<100000) then
       let
	  val d1 = canon t
	  val _ = dsay "\n# cannon\n"
	  val _ = dbg_out d1

	  val d1 = pl_shortcut d1
	  val d1 = delete_empty_context d1
	  val _ = dsay("\n# after pl_shortcut:\n")
	  val _ = dbg_out d1

	  val d1 = sharing_opt d1
	  val d1 = delete_empty_context d1
	  val _ = dsay("\n# after sharing:\n")
	  val _ = dbg_out d1

	  val d1 = color_opt d1
	  val d1 = delete_empty_context d1
	  val _ = dsay("\n# after color sharing:\n")
	  val _ = dbg_out d1

	  val d1 = size_opt d1
	  val d1 = delete_empty_context d1
	  val _ = dsay("\n# after size sharing:\n")
	  val _ = dbg_out d1

	  val d1 = toDoc d1
(*	  val d1 = RemoveRedundant.go d1*)

       in (T.size d1,d1) end
    else (sz,toDoc t)

    val d2 = combine_context t
    val d2 = pl_shortcut d2
    val d2 = delete_empty_context d2
    val _ = dsay("\n# after pl_shortcut:\n")
    val _ = dbg_out d2

    val d2 = sharing_opt d2
    val d2 = delete_empty_context d2
    val _ = dsay("\n# after sharing:\n")
    val _ = dbg_out d2

    val d2 = color_opt d2
    val d2 = delete_empty_context d2
    val _ = dsay("\n# after color sharing:\n")
    val _ = dbg_out d2

    val d2 = size_opt d2
    val d2 = delete_empty_context d2
    val _ = dsay("\n# after size sharing:\n")
    val _ = dbg_out d2

    val d2 = toDoc d2
(*    val d2 = RemoveRedundant.go d2*)

    val opt2 = T.size d2
  in if opt1>opt2 then d2 else d1
  end

(*  val big_test = Tests.readTest "test3"*)
  val tests = Tests.mk name true go
    ["<b>1</b><b>2</b><b>3</b><I>4</I><I>5</I>",
     "<r>aaa</r><g>ddd</g><b>ccc</b><g>ddd</g><r>eee</r>",
     "<B><I><r>0<g>1</g>2</r></I></B><m>3</m><c>4</c><y>5</y><c>7</c><m>9</m>",
     "<B><I><r>0<g>1</g>2</r></I></B><9>3</9><8>4</8><7>5</7><8>7</8><9>9</9>",
     "<B><PL><I>aaa<PL><I></B><EM><S><U>bbb</U></S></EM>",
     "<EM><B><I>aaa</I></B></EM><B>bbb</B><B><I><EM>aaa</EM></I><B>",
     "<I><S><TT>foo</TT></S></I> bar <TT><S><I> gee </I></S></TT><I><S><TT>foo</TT></S></I> bar <TT><S><I> gee </I></S></TT>",
     "<I><S><TT>foo</TT></S></I> bar <TT><S><I> gee </I></S></TT><I><S><TT>foo</TT></S></I>",
     "<I><S><TT>foo</TT></S></I><TT><S><I> gee </I></S></TT><I><S><TT>foo</TT></S></I><TT><S><I> gee </I></S></TT>",
     "<9><k><3>ddd</3>foo</k>ddd</9>",
     "<B><I>aaaa</I></B><I><B>bbbb</B></I>",
     "<g><U><0>bbb</0><k>bar</k>bar</U><S><y>iii</y><U>zzz</U>fff</S>foo</g>",
     "<U><U><U><U>foo</U>bar</U></U></U>"]

    val id_tests = Tests.mk "id-test" false (toDoc o fromDoc)
      ["1", "<B>1</B>", "<EM>1</EM>"]

end
